package job

import (
	"context"
	"flag"
	"time"

	"go.uber.org/zap"

	"github.com/freeman42332/zdb1000n/src/job/config"
	"github.com/freeman42332/zdb1000n/src/utils"
	"github.com/freeman42332/zdb1000n/src/utils/metrics"
)

// GlobalConfig passes commandline arguments to every job.
type GlobalConfig struct {
	ProxyURLs           string
	EnablePrimitiveJobs bool
	ScaleFactor         int
	MinInterval         time.Duration
	Backoff             utils.BackoffConfig
}

// NewGlobalConfigWithFlags returns a GlobalConfig initialized with command line flags.
func NewGlobalConfigWithFlags(log *zap.Logger) *GlobalConfig {
	res := GlobalConfig{}

	flag.StringVar(&res.ProxyURLs, "proxy", utils.GetEnvStringDefault("SYSTEM_PROXY", ""),
		"system proxy to set by default (can be a comma-separated list or a template)")
	flag.IntVar(&res.ScaleFactor, "scale", utils.GetEnvIntDefault("SCALE_FACTOR", 1, log),
		"used to scale the amount of jobs being launched, effect is similar to launching multiple instances at once")
	flag.DurationVar(&res.MinInterval, "min-interval", utils.GetEnvDurationDefault("MIN_INTERVAL", 0, log),
		"minimum interval between job iterations")

	flag.IntVar(&res.Backoff.Limit, "backoff-limit", utils.GetEnvIntDefault("BACKOFF_LIMIT", utils.DefaultBackoffConfig().Limit, log),
		"how much exponential backoff can be scaled")
	flag.IntVar(&res.Backoff.Multiplier, "backoff-multiplier", utils.GetEnvIntDefault("BACKOFF_MULTIPLIER", utils.DefaultBackoffConfig().Multiplier, log),
		"how much exponential backoff is scaled with each new error")
	flag.DurationVar(&res.Backoff.Timeout, "backoff-timeout", utils.GetEnvDurationDefault("BACKOFF_TIMEOUT", utils.DefaultBackoffConfig().Timeout, log),
		"initial exponential backoff timeout")

	return &res
}

// Job comment for linter
type Job = func(ctx context.Context, args config.Args, globalConfig *GlobalConfig, a *metrics.Accumulator, logger *zap.Logger) (data any, err error)

// Get job by type name
//nolint:cyclop // The string map alternative is orders of magnitude slower
func Get(t string) Job {
	switch t {
	case "http", "http-flood":
		return fastHTTPJob
	case "http-request":
		return singleRequestJob
	case "tcp":
		return tcpJob
	case "udp":
		return udpJob
	case "packetgen":
		return packetgenJob
	case "sequence":
		return sequenceJob
	case "parallel":
		return parallelJob
	case "log":
		return logJob
	case "set-value":
		return setVarJob
	case "check":
		return checkJob
	case "sleep":
		return sleepJob
	case "discard-error":
		return discardErrorJob
	case "timeout":
		return timeoutJob
	case "loop":
		return loopJob
	default:
		return nil
	}
}

type Config interface {
	FromGlobal(GlobalConfig)
}

func ParseConfig(c Config, args config.Args, global GlobalConfig) error {
	if err := utils.Decode(args, c); err != nil {
		return err
	}

	c.FromGlobal(global)

	return nil
}

// BasicJobConfig comment for linter
type BasicJobConfig struct {
	IntervalMs int
	Interval   *time.Duration
	utils.Counter
	Backoff *utils.BackoffConfig
}

func (c *BasicJobConfig) FromGlobal(global GlobalConfig) {
	if c.GetInterval() < global.MinInterval {
		c.Interval = &global.MinInterval
	}

	if c.Backoff == nil {
		c.Backoff = &global.Backoff
	}
}

func (c BasicJobConfig) GetInterval() time.Duration {
	return utils.NonNilOrDefault(c.Interval, time.Duration(c.IntervalMs)*time.Millisecond)
}

// Next comment for linter
func (c *BasicJobConfig) Next(ctx context.Context) bool {
	return utils.Sleep(ctx, c.GetInterval()) && c.Counter.Next()
}
