package metrics

import (
	"fmt"
	"io"
	"sort"
	"sync"
	"text/tabwriter"
)

// Stat is the type of statistical metrics.
type Stat int

const (
	RequestsAttemptedStat Stat = iota
	RequestsSentStat
	ResponsesReceivedStat
	BytesSentStat

	NumStats
)

// String representation of the Stat.
func (s Stat) String() string {
	return [...]string{
		"requests attempted",
		"requests sent",
		"responses received",
		"bytes sent",
	}[s]
}

type dimensions struct {
	jobID  string
	target string
}

// Reporter gathers metrics across jobs and reports them.
// Concurrency-safe.
type Reporter struct {
	metrics [NumStats]sync.Map // Array of metrics by Stat. Each metric is a map of uint64 values by dimensions.
}

// NewReporter creates a new Reporter.
func NewReporter() *Reporter { return &Reporter{} }

// Sum returns a total sum of metric s.
func (r *Reporter) Sum(s Stat) uint64 {
	var res uint64

	r.metrics[s].Range(func(_, v interface{}) bool {
		value, ok := v.(uint64)
		if !ok {
			return true
		}

		res += value

		return true
	})

	return res
}

type (
	// Stats contains all metrics packed as an array.
	Stats [NumStats]uint64
	// MultiStats contains multiple Stats as a slice. Useful for collecting Stats from coroutines.
	MultiStats []Stats
	// PerTargetStats is a map of Stats per target.
	PerTargetStats map[string]Stats
)

// NewStats returns a new Stats record.
func NewStats(requestsAttempted, requestsSent, responsesReceived, bytesSent uint64) Stats {
	return Stats{requestsAttempted, requestsSent, responsesReceived, bytesSent}
}

// Sum up all Stats into a total Stats record.
func (s MultiStats) Sum() Stats {
	var res Stats

	for i := range s {
		for j := RequestsAttemptedStat; j < NumStats; j++ {
			res[j] += s[i][j]
		}
	}

	return res
}

func (ts PerTargetStats) sum(s Stat) uint64 {
	var res uint64

	for _, v := range ts {
		res += v[s]
	}

	return res
}

func (ts PerTargetStats) sortedTargets() []string {
	res := make([]string, 0, len(ts))
	for k := range ts {
		res = append(res, k)
	}

	sort.Strings(res)

	return res
}

// SumAllStatsByTarget returns a total sum of all metrics by target.
func (r *Reporter) SumAllStatsByTarget() PerTargetStats {
	res := make(PerTargetStats)

	for s := RequestsAttemptedStat; s < NumStats; s++ {
		r.metrics[s].Range(func(k, v interface{}) bool {
			d, ok := k.(dimensions)
			if !ok {
				return true
			}

			value, ok := v.(uint64)
			if !ok {
				return true
			}

			stats := res[d.target]
			stats[s] += value
			res[d.target] = stats

			return true
		})
	}

	return res
}

// WriteSummary dumps Reporter contents into the target.
func (r *Reporter) WriteSummary(target io.Writer) {
	w := tabwriter.NewWriter(target, 1, 1, 1, ' ', tabwriter.AlignRight)

	defer w.Flush()

	stats := r.SumAllStatsByTarget()

	fmt.Fprintln(w, "\n --- Traffic stats ---")
	fmt.Fprintf(w, "|\tTarget\t|\tRequests attempted\t|\tRequests sent\t|\tResponses received\t|\tData sent \t|\n")

	const BytesInMegabyte = 1024 * 1024

	var totals Stats

	for _, tgt := range stats.sortedTargets() {
		tgtStats := stats[tgt]

		fmt.Fprintf(w, "|\t%s\t|\t%d\t|\t%d\t|\t%d\t|\t%.2f MB \t|\n", tgt,
			tgtStats[RequestsAttemptedStat],
			tgtStats[RequestsSentStat],
			tgtStats[ResponsesReceivedStat],
			float64(tgtStats[BytesSentStat])/BytesInMegabyte,
		)

		for s := range totals {
			totals[s] += tgtStats[s]
		}
	}

	fmt.Fprintln(w, "|\t---\t|\t---\t|\t---\t|\t---\t|\t--- \t|")
	fmt.Fprintf(w, "|\tTotal\t|\t%d\t|\t%d\t|\t%d\t|\t%.2f MB \t|\n\n",
		totals[RequestsAttemptedStat],
		totals[RequestsSentStat],
		totals[ResponsesReceivedStat],
		float64(totals[BytesSentStat])/BytesInMegabyte,
	)
}

// NewAccumulator returns a new metrics Accumulator for the Reporter.
func (r *Reporter) NewAccumulator(jobID string) *Accumulator { return newAccumulator(jobID, r) }

// Accumulator for statistical metrics for use in a single job. Requires Flush()-ing to Reporter.
// Not concurrency-safe.
type Accumulator struct {
	jobID   string
	metrics [NumStats]map[string]uint64 // Array of metrics by Stat. Each metric is a map of uint64 values by target.

	r *Reporter
}

// Add n to the Accumulator Stat value. Returns self for chaining.
func (a *Accumulator) Add(target string, s Stat, n uint64) *Accumulator {
	a.metrics[s][target] += n

	return a
}

// Inc increases Accumulator Stat value by 1. Returns self for chaining.
func (a *Accumulator) Inc(target string, s Stat) *Accumulator { return a.Add(target, s, 1) }

// AddStats to the Accumulator. Returns self for chaining.
func (a *Accumulator) AddStats(target string, s Stats) *Accumulator {
	for i := range a.metrics {
		a.metrics[i][target] += s[i]
	}

	return a
}

// Flush Accumulator contents to the Reporter.
func (a *Accumulator) Flush() {
	for stat := RequestsAttemptedStat; stat < NumStats; stat++ {
		for target, value := range a.metrics[stat] {
			a.r.metrics[stat].Store(dimensions{jobID: a.jobID, target: target}, value)
		}
	}
}

// Clone a new, blank metrics Accumulator with the same Reporter as the original.
func (a *Accumulator) Clone(jobID string) *Accumulator { return newAccumulator(jobID, a.r) }

func newAccumulator(jobID string, r *Reporter) *Accumulator {
	res := &Accumulator{
		jobID: jobID,
		r:     r,
	}

	for s := RequestsAttemptedStat; s < NumStats; s++ {
		res.metrics[s] = make(map[string]uint64)
	}

	return res
}

// NopWriter implements io.Writer interface to simply track how much data has to be serialized
type NopWriter struct{}

func (w NopWriter) Write(p []byte) (int, error) { return len(p), nil }
